import os
import sys


def test_import():
    import vipy
    import vipy.activity
    import vipy.annotation
    import vipy.calibration
    import vipy.camera
    import vipy.collector
    import vipy.downloader
    import vipy.dropbox
    import vipy.geometry
    import vipy.image
    import vipy.linalg
    import vipy.math
    import vipy.metrics
    import vipy.object
    import vipy.show
    import vipy.ssim
    import vipy.util
    import vipy.video
    import vipy.videosearch
    import vipy.visualize
    import vipy.gui.using_matplotlib
    import vipy.dataset.aflw
    import vipy.dataset.caltech256
    import vipy.dataset.ethzshapes
    import vipy.dataset.fddb
    import vipy.dataset.imagenet
    import vipy.dataset.kthactions
    import vipy.dataset.megaface
    import vipy.dataset.msceleb
    import vipy.dataset.vggface
    import vipy.dataset.youtubefaces
    import vipy.dataset.activitynet
    import vipy.dataset.caltech101
    import vipy.dataset.casia
    import vipy.dataset.facescrub
    import vipy.dataset.hmdb
    import vipy.dataset.kinetics
    import vipy.dataset.lfw
    import vipy.dataset.mnist
    import vipy.dataset.vggface2
    print('[test_import]: PASSED')


if __name__ == "__main__":
    test_import()
